package ro.uvt.cs.ads;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *  The {@code BPQ} class represents a bounded priority queue
 *  with a bounded capacity k which stores k elements with 
 *  the smallest priorities from a collection of n elements.
 * @author Mircea Marin
 *
 */
public class BPQ<Item> implements Iterable<Item> {
	private int size;    // number of elements stored in this queue
	private double[] p;  // priorities
	private Item[] item; // items 

	public BPQ(int k) {
		size = 0;
		p = new double[k];
		item = (Item[]) new Object[k];
	}
	
	/**
	 * Checks if this queue is full.
	 * @return {@code true} if this queue is full, and {@code false} otherwise. 
	 */
	public boolean full() {
		return size == p.length;
	}

	public double getMaxPriority() {
		return p[size - 1];
	}
	
	/**
	 * Returns the size of this queue.
	 * @return the size of this queue
	 */
	public int size() {
		return size;
	}
	
	/**
	 * Returns the capacity of this queue.
	 * @return the capacity of this queue
	 */
	public int capacity() {
		return p.length;
	}
	
	/**
	 * Inserts element {@code e} with priority {@code pr} in this queue.
	 * @param e the element
	 * @param pr the priority of {@code e}
	 */
	public void insert(Item e, double pr) {
		int k = p.length;
		if (size==0||p[size-1]<=pr) {
			if (size<k) {
				item[size]=e;
				p[size++]=pr;
			}
		} else { // size>0 && p[size-1]>p
			int c = size-1;
			if (size<k) {
				item[size]=item[c];
				p[size]=p[c];
				size++;
			}
			while ((c>0) && (p[c-1]>pr)) {
				item[c]=item[c-1];
				p[c]=p[c-1];
				c--;
			}
			item[c]=e;
			p[c]=pr;
		}	
	}

	public String toString() {
		String r = "";
		for (int i=0;i<size;i++) 
			r=r+item[i].toString()+"["+p[i]+"]:";
		return r;
	}
	
	/**
	 * Unit test
	 * @param args
	 */
	public static void main(String[] args) {
		BPQ<String> bpq = new BPQ<String>(4);
		bpq.insert("A",1.);
		bpq.insert("B",3.);
		bpq.insert("C",2.);
		bpq.insert("D",4.);
		System.out.println(bpq.toString());
		bpq.insert("X",.5);
		System.out.println(bpq.toString());
		bpq.insert("Y",1.5);
		System.out.println(bpq.toString());
	}

	@Override
	public Iterator<Item> iterator() {
		return new PointIterator();
	}

	// an iterator for the points in this class
	private class PointIterator implements Iterator<Item> {
		int current;
		
		public PointIterator() {
			current =0;
		}
		
		public boolean hasNext() {
            return current < size;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }

        public Item next() {
            if (!hasNext()) throw new NoSuchElementException();
            return item[current++];
        }
	}
}
