package ro.uvt.cs.ads;

import java.util.Arrays;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;

/**
 * The {@code Node} class represents a node in a k-d tree of points with <tt>double</tt> coordinates.
 *
 */
public class Node {
	private double[] x;   // element
	private Node l;       // left subtree
	private Node r;       // right subtree
	
	/**
	 * Constructs a node for point {@code y}.
	 * @param y
	 */
	public Node(double[] y) {
		x=new double[y.length];
		for (int i=0;i<x.length;i++) 
			x[i]=y[i];	
	}
	
	/**
	 * Returns the dimension of the point in this node.
	 * @return the dimension of the point in this node.
	 */
	public int dimension() {
		return x.length;
	}
		
	public void makeTree(char side, double[][] points,int s,int e,int axis) {
		if(e<=s) return;
		if (e==s+1) {
			if (side=='l') l=new Node(points[s]);
			else r=new Node(points[s]);
		} else {
			Arrays.sort(points,s,e,
					(x,y) -> (x[axis]<y[axis])?-1:(x[axis]==y[axis]?0:1));
			int m = (e+s)/2;
			while ((m+1<e)&&points[m][axis]==points[m+1][axis]) m++;
			Node crt = (side=='l')?(l=new Node(points[m])):(r=new Node(points[m]));
			crt.makeTree('l',points,s,  m,(axis+1) % dimension());
			crt.makeTree('r',points,m+1,e,(axis+1) % dimension());
		}
	}
	
	private static double[] testPoint, guess;
	private static double bestDistance; 
	
	/**
	 * Returns the nearest neighbor to the test point {@code test} 
	 * in the k-d tree whose root is {@code crt}
	 *  
	 * @param crt the current node
	 * @param test the test point
	 * @return the nearest neighbor
	 */
	protected static double[] NN1(Node crt,double[] test) {
		testPoint = test;
		bestDistance = Double.POSITIVE_INFINITY;
		guess = null;		
		nearest1(crt,0);
		return guess;
	}
	
	private static void nearest1(Node crt, int axis) {
		if (crt==null) return;
		double d = dist(testPoint,crt.x);
		if (d<bestDistance) {
			bestDistance = d;
			guess = crt.x;
		}
		// Recursively search the half of the tree that contains the test point
		char side = (testPoint[axis]<=crt.x[axis])?'l':'r';
		nearest1((side=='l')?crt.l:crt.r,(axis+1)%crt.dimension());
		// Perform branch-and-bound search
		if (Math.abs(crt.x[axis]-testPoint[axis])<bestDistance)
			/*  search the other subtree on the next axis */
			nearest1((side=='l')?crt.r:crt.l,(axis+1)%crt.dimension());
	}

	protected static BPQ<double[]> NNn(Node crt, double[] test, int k) {
		BPQ<double[]> bpq = new BPQ<>(k);

		// All variables referenced from lambdas must be final
		// Therefore I defined the RecursiveLambda type which will
		// contain the lambda in order for the lambda to be recursive
		//
		// I decided to use a lambda instead of creating a separate function in
		// order to not need to declare bpq as a global
		final RecursiveLambda<BiConsumer<Node, Integer>> rl = new RecursiveLambda<>();
		rl.func = (node, axis) -> {
			if (node == null) {
				return;
			}

			bpq.insert(node.x, dist(node.x, test));

			char side;
			if (test[axis] < node.x[axis]) {
				side = 'l';
			}
			else {
				side = 'r';
			}
			rl.func.accept((side=='l')?crt.l:crt.r,(axis+1)%crt.dimension());

			if (!bpq.full() || Math.abs(node.x[axis] - test[axis]) < bpq.getMaxPriority()) {
				rl.func.accept((side=='l')?crt.r:crt.l,(axis+1)%crt.dimension());
			}
		};

		rl.func.accept(crt, 0);

		return bpq;
	}
		
	protected void toString(StringBuffer sb, String indent) {
		sb.append(indent).append(stringVal());
		if (l!=null) { sb.append('\n'); l.toString(sb,indent+" "); }
		if (r!=null) { sb.append('\n'); r.toString(sb,indent+" "); }
	}
	
	/**
	 * The string representation of the point in this node
	 * @return the string representation of the point in this node
	 */
	private String stringVal() {
		StringBuffer sb=new StringBuffer();
		sb.append('(');
		int n = x.length;
		for (int i=0;i<n-1;i++)
			sb.append(String.format("%5.2f,", x[i]));
		sb.append(String.format("%5.2f)",x[n-1]));
		return sb.toString();
	}
	
	private void validateNode(double[] y) {
		if (y.length!=x.length) 
			throw new IllegalArgumentException("Illegal point: wrong dimension.");
	}
	
	public static double dist(double[] x, double[] y) {
		if(x==null || y==null || x.length!=y.length) 
			throw new IllegalArgumentException("Illegal arguments:"+
		      "they should be non-null instances of double[] of same size.");
		double result=0.0;
		for (int i=0;i<x.length;i++) 
			result += Math.pow(x[i]-y[i],2);
		return Math.sqrt(result);
	}

}
